#! /usr/local/bin/python2.4
#Lauren Perskie
#7/17/07
#This program calculates hbond satisfaction in a protein
#See below for more detailed comments
#Last update: 8/30/07
#
#Notes: The N-Terminus of the protein is not checked for hbond satisfaction
#       The flow of this program with LINUS is as follows:
#       0)  Set hbondParams options in linus.py using set_hbs_parameters
#       1)  LINUS does a move and checks the structure for steric clash
#       2)  If there are no clashes, the structure is passed into the
#           hbondSatisfied function.
#       3)  The length of segment of protein that is checked is related to
#           how advanced the simulation is.  Earlier stages of the simulation
#           check shorter segments of the protein.  The segments sizes are
#           determined by moveset+(numres/numcycles)*current_cycle_number
#           A 100 residue protein running 500 cycles would check 3 residues
#           in the first ~5 iterations
#       4)  Quick check is used only after an hbond-satisfied protein is found
#       5)  The program checks the protein segment for bb-bb, water-bb, and
#           sc-bb hbonds.  If all hbonds IN THE SEGMENT are satisfied,
#           hbondSatisfied returns 1 to LINUS, and LINUS accepts the structure.
#           If there is even 1 unsatisfied hbond, hbondSatisfied returns 0, and
#           LINUS rejects the structure.

import hbond, Atom3d, linusMol, pdbio, sys, string
from sort_sc import sort_sc_pos

import Numeric as N

#Names of bb donors and acceptors
bb_donor    = [' N  '];
bb_acceptor = [' O  '];

#Names of all donors and acceptors
all_donor   = [' N  ',' NZ ', ' ND1',' ND2', ' NE2',' NH1',' NH2'];
all_acceptor= [' O  ',' OD1', ' OD2',' OE1', ' OE2',' OG1',' OH '];

#Residue names for sc donors, sc acceptors, and scs that serve as both
sc_donors    = ['LYS','ARG','HIS'];
sc_acceptors = ['ASP','GLU','ASN'];
sc_both = ['SER','THR','ASN','GLN'];

#Distances, azimuthal & altitudinal angles for water placement for both
#NHs and COs
dDists = {'short':[1.97]*5,'long':[1.97]*5+[2.45]*5};
aDists = {'short':[2.95]*5,'long':[2.95]*5+[3.45]*5};
dAngs  = {'short':[180.0]+[166.0]*4, \
          'long' :[180.0]+[166.0]*4+[180.0]+[166.0]*4};
aAngs  = {'short':[180.0]+[130.0]*4,\
          'long' :[180.0]+[130.0]*4+[180.0]+[130.0]*4}
dTors  = {'short':[0.0,180.0,90.0,0.0,-90.0],\
          'long' :[0.0]+range(-180,180,90)+[0.0]+range(-180,180,90)};
aTors  = {'short':[0.0,180.0,90.0,0.0,-90.0],\
          'long' :[0.0]+range(-180,180,90)+[0.0]+range(-180,180,90)};

hbondParams  = {'dDists':dDists['short'],#distances between waters & H
                'dAngles':dAngs['short'],#altitudinal ang btw waters, H, & N 
                'dTorsions':dTors['short'],#azimuthal ang btw wat, H, N & C
                'aDists':aDists['short'],#distances between waters and O
                'aAngles':aAngs['short'],#altitudinal ang btw waters, O & C  
                'aTorsions':aTors['short'],#azimuthal ang btw wat, O, C & CA
                'check_entropy':1, #Count how many waters satisfy hbond (1)
                                   #Or stop after finding 1 (0)
                'bbTheta1':90.0,   #Min Angle between CA,N,O
                'bbTheta2':90.0,   #Min Angle between C,O,N
                'bbTorsion':-999.9,#Torsion angle for hbonding (CA,N,O,C)
                'bbDistance':3.5,  #Max Distance between bb donor and acceptor
                'scTheta1':90.0,   #Min SC angle 1
                'scTheta2':90.0,   #Min SC angle 2
                'scTorsion':-999.9,#Min sc torsion
                'scDistance':3.5,  #Min sc donor/acceptor distance
                'd_types':bb_donor,#Types of donors
                'a_types':bb_acceptor,#Types of acceptors
                'water_radius':1.4,#Radius of water probe
                'print_results': 0,#print results
                'mode':'bb_only',  #Hbond checking mode
                'wat_mask':-1000000, #Value used to "mask" waters
                'sc_mask':100000, #Value used to "mask" sidechains
                'wfilename':''}#Name of water outfilename

hbondLists = {'donors':None,
              'acceptors':None,
              'tmpDonorSat':None,#Temporary lists
              'tmpAcceptorSat': None,#for how hbonds are satisfied
              'donorSatTypes':None,  #(numerical and 
              'acceptorSatTypes':None};#word codes)




#Maximum distance between sidechain CA and donor/acceptor

scParams = {'ARG':20.8, 'ASP': 12.9, 'ASN':12.9, 'GLN': 16.3, 'GLU':16.3,
            'LYS':16.7, 'HIS':12.9, 'SER':6.5, 'THR':6.5};


def hbondInit(mol,opts,hbLists):

    hbLists['donors'], hbLists['acceptors'] = \
                        get_hbond_atoms(mol,opts['mode']);
    return

def atoms_with_names(atoms,names):

    return [x for x in xrange(len(atoms)) if (atoms[x].name in names)];

def atoms_in_residues(mol,atomList,residue_names):

    return [x for x in atomList if \
            (mol.residue_names[mol.atoms[x].resnum] in residue_names)];


#Satisfaction is determined by whether the donor_sat and acceptor_sat
#lists do not contain -1
#(-1 is the code for an hbond being unsatisfied)

def satisfied(satlist):

    return [x for x in satlist if x == -1] == []



#Remove proline N's from the donor_list.  This is NOT used in LINUS simulations
#Linus simulations mark proline Ns in the list by -1000000 and move on.
#This allows donor_sat and acceptor_sat lists to be indexed by their
#residue numbers

def remove_pro_n(mol,donor_list):

    pro_fai = [mol.residue_first_atom_indices[x] \
               for x in xrange(len(mol.residue_names)) \
               if mol.residue_names[x] == 'PRO'];
    
    #Since N is always at the first atom index, it suffices to pull
    #the Ns that match PRO first atom indices out
    
    return [x for x in donor_list if not (x in pro_fai)];



#Sets all proline Ns in donor_sat list equal to -1000001
#This allows donor_sat and acceptor_sat lists to be indexed by
#their residue numbers

def satisfy_donor_pros(mol,donors,donor_sat):

    for i in xrange(len(donors)):

        if mol.residue_names[mol.atoms[donors[i]].resnum] == 'PRO':
            donor_sat[i] = -1000001;
            
    return



#Get Hbond Donor and Acceptor lists (donor_list and acceptor_list)
#LINUS uses bb_only (since those are the only hbonds we're concerned with)
#but bb+sc can be used

def get_hbond_atoms(mol,mode='bb_only'):

    donors = [];
    acceptors = [];

    if mode == 'bb_only':
        donors    = atoms_with_names(mol.atoms,bb_donor);
        acceptors = atoms_with_names(mol.atoms,bb_acceptor);
    else:
        donors    = atoms_with_names(mol.atoms,all_donor);
        acceptors = atoms_with_names(mol.atoms,all_acceptor);

    return [donors,acceptors];


###############################################################################
#hbondSatisfied                                                               #
#                                                                             #
#Actual program LINUS calls to check hbond satisfaction of a                  #
#segment of the protein                                                       #
#Inputs:                                                                      #
#       mol: the protein (a LinusMol)                                         #
#       st:  the starting residue to check                                    #
#       fh:  the last residue to check+1                                      #
#            (so that the slices will terminate at the last residue)          #
#       opts:dictionary of hbondParams.                                       #
#            User sets to specific values and passes it to fxn                #
#                                                                             #
#Outputs:                                                                     #
#       a Boolean value specifying whether all hbonds are satisfied (True)    #
#       or not satisfied (False)                                              #
#                                                                             #
#Description:                                                                 #
#                                                                             #
#       This function checks hydrogen bonds in three ways.                    #
#       1)  Looks for intrapeptide backbone-backbone hbonds                   #
#       2)  Looks for backbone-water hbonds                                   #
#       3)  Looks for backbone-sidechain hbonds                               #
#                                                                             #
#       If all hbonds in the segment are satisfied, the way they are satisfied#
#       is recorded in temporary lists in hbondParams.  That way previously   #
#       satisfied hbonds can be retested after the next move.  This is meant  #
#       to save time in later stages when many hbonds are checked, but only   #
#       a few hbonds are affected by a move.                                  #
#       The quick_check function does these checks on both donors and         #
#       acceptors                                                             #
#                                                                             #
#       The function then returns a boolean value indicating whether          #
#       all hbonds are satisfied or not.                                      #
#                                                                             #
###############################################################################

def hbondSatisfied(mol,st,fh,iteration,opts=hbondParams,hbls=hbondLists,
                   writepdb=0):    

    #Set options and initialize variables
    
    donors,acceptors = [hbls['donors'],hbls['acceptors']];
    donor_sat     = [-1]*len(donors);   # -1 is the sat code for unsatisfied
    donor_types   = ['NS']*len(donors); #'NS' is the types code for unsatisfied
    acceptor_sat  = [-1]*len(acceptors);
    acceptor_types= ['NS']*len(acceptors);
    
    #Make sure starting index >= 0
    
    st = max(0,st);
    
    #Make sure ending index <= num_residues
    
    fh = min(mol.num_residues,fh);

    #If protein has not been previously hbond satisfied, temp lists will not
    #exist, and quick_check cannot be used, so skip it
    #If protein has been previously hbond satisfied, see if any donors or
    #acceptors are satisfied in the same way
    
    if hbls['tmpDonorSat'] != None:

        hbond.quick_check(mol.atoms,donors,donor_sat,
                          hbls['tmpDonorSat'],
                          donor_types,hbls['donorSatTypes'],
                          mol.residue_names,
                          scParams,opts['dDists'],opts['dAngles'],
                          opts['dTorsions'],
                          mol.residue_first_atom_indices,st,fh,
                          opts['check_entropy'],fh-st+1,
                          opts['bbDistance'],
                          opts['bbTheta1'],opts['bbTheta2'],opts['bbTorsion'],
                          opts['water_radius'],opts['wat_mask']);

        hbond.quick_check(mol.atoms,acceptors,acceptor_sat,
                          hbls['tmpAcceptorSat'],
                          acceptor_types,hbls['acceptorSatTypes'],
                          mol.residue_names,
                          scParams,opts['aDists'],opts['aAngles'],
                          opts['aTorsions'],
                          mol.residue_first_atom_indices,st,fh,
                          opts['check_entropy'],fh-st+1,
                          opts['bbDistance'],
                          opts['bbTheta1'],opts['bbTheta2'],opts['bbTorsion'],
                          opts['water_radius'],opts['wat_mask']);
        
    #Set SC values after quick check so that sidechains will not be reused
    scSat = 0;
    scPartners,donStart,accEnd = \
                             sc_donor_acceptor_CAs(mol,donor_sat,acceptor_sat,
                                                   st,fh);
    scPartnerNames = get_sc_names(mol,scPartners);

    #Mark donor_pros and N-terminus so that they are not checked
    satisfy_donor_pros(mol,donors,donor_sat);
    donor_sat[0] = -1000002;

    #Do the 3 checks for hbond satisfaction
    
    hbond.intra_peptide_hbonds(mol.atoms,donors[st:fh],acceptors[st:fh],
                               donor_sat,acceptor_sat, donor_types,
                               acceptor_types, fh-st+1,
                               opts['bbDistance'],opts['bbTheta1'],
                               opts['bbTheta2'],opts['bbTorsion']);
    
    hbond.water_hbonds(mol.atoms,donors[st:fh],acceptors[st:fh],
                       donor_sat,acceptor_sat,donor_types,acceptor_types,
                       mol.residue_first_atom_indices,
                       opts['dDists'],opts['dAngles'],opts['dTorsions'],
                       opts['aDists'],
                       opts['aAngles'],opts['aTorsions'],opts['check_entropy'],
                       fh-st+1,opts['water_radius'],opts['wat_mask']);

    #If bb hbonds are all satisfied by bb and water, skip sc check
    
    if not (satisfied(donor_sat[st:fh]) and satisfied(acceptor_sat[st:fh])):
        scSat = hbond.bb_sc_sat(mol.atoms,donors[st:fh],acceptors[st:fh],
                                donor_sat,acceptor_sat,scPartners,
                                scPartnerNames,
                                scParams,accEnd,donStart,fh-st+1,
                                opts['sc_mask']);
        get_sc_pos(donor_sat,acceptor_sat,scPartners,donor_types,
                   acceptor_types,opts['sc_mask']);



    #If segment is hbond satisfied, record the type of satisfaction
    #and its corresponding numeric code in temp lists so that quick_check
    #can be done after next move.
    
    if satisfied(donor_sat[st:fh]) and satisfied(acceptor_sat[st:fh]):
        hbls['tmpDonorSat'] = donor_sat;
        hbls['tmpAcceptorSat'] = acceptor_sat;
        hbls['donorSatTypes'] = donor_types;
        hbls['acceptorSatTypes'] = acceptor_types;

    if writepdb:
       pdbWithWaters(mol,donors,acceptors,donor_sat,acceptor_sat,opts)
       
    #Return boolean value for hbond satisfaction
    return (satisfied(donor_sat[st:fh]) and satisfied(acceptor_sat[st:fh]));

#Convert decimal numbers representing water positions into binary lists
#of 1s and 0s

def get_water_pos(nums,npos):

    return [[(n/x)%2 for x in [2**y for y in range(npos)]] for n in nums];

#Put waters into the protein file.  This is used to write a pdb file with
#waters.  If check_entropy=1, only the waters that satisfy are placed.
#If check_entropy=0, all waters are placed

def place_waters(atomIdxs,mol,sat,dists,angles,torsions,
                 check_entropy,wat_mask):

    m = 0;
    n = 1;
    new = Atom3d.Atom3d();
    ma = mol.atoms;
    H = Atom3d.Atom3d();

    watPos = [0]*len(sat);
    for i in xrange(len(sat)):
        if wat_mask < sat[i] < -1:
            watPos[i] = sat[i];

    if check_entropy:
        binWatPos = get_water_pos(watPos,len(dists));
    else:
        binWatPos = [[1]*len(dists)]*len(watPos);
    
    for i in xrange(len(atomIdxs)):
        if sum(binWatPos[i]) == 0:
            continue

        if mol.atoms[atomIdxs[i]].name == ' N  ':
            H = Atom3d.Atom3d();
            H.resnum = ma[atomIdxs[i]].resnum;
            if ma[atomIdxs[i]].resnum != 0:
                hbond.make_atom(ma[atomIdxs[i]],1.008,
                                ma[atomIdxs[i]].first_parent,119.15,
                                ma[atomIdxs[i]].second_parent,0.0,H);
            else:
                continue

        m = 0;
        
        for j in xrange(len(dists)):
            if binWatPos[i][j] == 1:
                new = Atom3d.Atom3d();
                new.name = ' O  ';
                if ma[atomIdxs[i]].resnum == 0 \
                       and ma[atomIdxs[i]].name == ' N  ':
                    hbond.make_atom(H,dists[j],
                                    ma[atomIdxs[i]],angles[j],
                                    ma[atomIdxs[i]+1],torsions[j],new);
                elif ma[atomIdxs[i]].name == ' N  ':
                    hbond.make_atom(H,dists[j],
                                    ma[atomIdxs[i]],angles[j],
                                    ma[atomIdxs[i]].first_parent,
                                    torsions[j],new);

                else:
                    hbond.make_atom(ma[atomIdxs[i]],dists[j],
                                    ma[atomIdxs[i]].first_parent,angles[j],
                                    ma[atomIdxs[i]].second_parent,
                                    torsions[j],new);
                
                mol.residue_names.append('WAT');
                mol.num_residues = mol.num_residues+1
                mol.res_pdb_number.append('  '+str(mol.num_residues).ljust(2))
                ma.append(new);
                mol.residue_first_atom_indices.append(mol.residue_first_atom_indices[-1]+1);

#Get indices for ALL sc donor and and acceptor CAs

def get_sc_donor_acceptor_CAs(mol):

    CAs = atoms_with_names(mol.atoms,' CA ')
    scAcc = atoms_in_residues(mol,CAs,sc_acceptors);
    scBot = atoms_in_residues(mol,CAs,sc_both);
    scDon = atoms_in_residues(mol,CAs,sc_donors);

    return [scAcc+scBot+scDon,len(scAcc),len(scAcc+scDon)-1];

#Get indices for all UNUSED sc donor and acceptor CAs

def sc_donor_acceptor_CAs(mol,donor_sat,acceptor_sat,startidx,endidx):

    CAs = atoms_with_names(mol.atoms,' CA ')
    scAcc = [x for x in atoms_in_residues(mol,CAs,sc_acceptors) \
             if x not in donor_sat and \
             startidx<= mol.atoms[x].resnum<= endidx];
    scBot = [x for x in atoms_in_residues(mol,CAs,sc_both) \
             if x not in donor_sat+acceptor_sat and \
             startidx<= mol.atoms[x].resnum <= endidx];
    scDon = [x for x in atoms_in_residues(mol,CAs,sc_donors) \
             if x not in acceptor_sat and \
             startidx<=mol.atoms[x].resnum<=endidx];             

    return [scAcc+scBot+scDon,len(scAcc),len(scAcc+scDon)-1];


#Get the residue names corresponding to the CA idxs

def get_sc_names(mol,idxs):

    return [mol.residue_names[mol.atoms[x].resnum] for x in idxs];

#Assign bb donors/acceptors to the best possible sidechain
def get_sc_pos(donor_sat,acceptor_sat,scPartners,donor_types,acceptor_types,
               sc_mask):

    #If there are no sc partners, return
    if scPartners == []:
        return

    #Make a numeric array of all side chain indices satisfying an hbond
    #scIdxs is a 3xn list, where N is the number of residues satisfied by a
    #side chain.
    #scIdxs[0] = decimal number representing the sc positions that satisfy
    #scIdxs[1] = donor_sat idx corresponding to scIdxs[0]
    #scIdxs[2] = boolean value.  0->donor,1->acceptor

    
    scIdxs = N.array([[donor_sat[x],x,0] for x in xrange(len(donor_sat))
                      if donor_sat[x] > sc_mask]+\
                      [[acceptor_sat[x],x,1] for x in xrange(len(acceptor_sat))
                      if acceptor_sat[x] > sc_mask]);

    #If there are no sidechain hbonds, return

    if len(scIdxs) == 0:
        return

    #A matrix of 1s and 0s corresponding to sc positions that satisfy
    #or do not satisfy each donor/acceptor satisfied by a sc
    
    sc_pos = N.array(get_water_pos(scIdxs[:,0]-sc_mask,len(scPartners)));
    

    #Simplify the matrix so that each donor/acceptor has its own unique
    #sidechain or get as close as possible.
    
    satisfied = sort_sc_pos(sc_pos);

    #If each donor/acceptor can have its own unique sc, it's satisfied
    
    if satisfied:

        scOrder = [[x for x in xrange(len(sc_pos[y])) if sc_pos[y,x] == 1][0]\
                   for y in xrange(len(sc_pos)) if sum(sc_pos[y])==1]

    #Else, match side chains as best you can.
    
    else:
        
        #Unmatched sidechains have value -999
        
        scOrder = [-999]*len(sc_pos);

        #Matched sidechains get the sc idx corresponding to the index holding
        #their atom index information.
        #Ex: scOrder[i] = 5.  Then scPartners[scOrder[i]] = 31
        #where 31 is the atom index of the 6th CA.

        for i in xrange(len(sc_pos)):
            if sum(sc_pos[i]) == 1:
                for j in xrange(len(sc_pos[i])):
                    if sc_pos[i,j] == 1:
                        scOrder[i] = j;
                        break

    #Match donor_sat and acceptor_sat to their appropriate side chains
    #These lists are passed by reference, so the changes will be remembered.
    
    for i in xrange(len(scIdxs)):

        if scOrder[i] == -999:
            if scIdxs[i,2] == 0:
                donor_sat[scIdxs[i,1]] = -1;
            else:
                acceptor_sat[scIdxs[i,1]] = -1;
            continue

        if scIdxs[i,2] == 0:
            donor_sat[scIdxs[i,1]] = scPartners[scOrder[i]];
            donor_types[scIdxs[i,1]] = 'SC'

        else:
            acceptor_sat[scIdxs[i,1]] = scPartners[scOrder[i]];
            acceptor_types[scIdxs[i,1]] = 'SC'
            

#Convert water number to binary and sum up the 1s.
#This corresponds to the number of positions water could assume
#without a steric clash

def get_num_waters(n,npos):

    return sum([(n/x)%2 for x in [2**y for y in range(npos)]]);


def pdbWithWaters(mol,donors,acceptors,donor_sat,acceptor_sat,opts):

    place_waters(acceptors,mol,acceptor_sat,opts['aDists'],
                 opts['aAngles'],opts['aTorsions'],opts['check_entropy'],
                 opts['wat_mask']);
    place_waters(donors,mol,donor_sat,opts['dDists'],opts['dAngles'],
                 opts['dTorsions'],opts['check_entropy'],opts['wat_mask']);
    pdbio.write_pdb(mol,opts['wfilename']+'.pdb');
                
            

        




        

    

        
    

    

    
            


    
                
                        
                    
                

    
    

    


    

    
        
            
            


