"""Module that defines the LinusProtein class"""

import Atom3d
import pdbio

class LinusProtein:
    """Class to describe a protein for simulation in LINUS.

    Instantiation

       Instantiation of this class requires one argument - the
       name of the file containing a description of the molecule
       in PDB format.

    Class Attributes

       o *filename* - string - name of file from which the molecule
       information was read

       o *atoms* - list - list of atoms in the molecule. each item
       in this list is an instance of *Atom3d*

       o *num_atoms* - integer - number of atoms in molecule

       o *num_residues* - integer - number of residues in molecule

       o *residue_names* - list - name of each residue in molecule

       o *residue_first_atom_indices* - list - index of the the
       first atom of each residue (this is horrendously named)

       o *phi_atoms* - list - contains a reference to the 'C' atom
       of each residue

       o *psi_atoms* - list - contains a reference to the 'O' atom
       of each residue

       o *omega_atoms* - list - contains a reference to the 'CA' atom
       of each residue

       o *chi_torsions* - list of lists - references to the atoms
       corresponding to the various chi torsions of each residue

       o *residue_helix_weights* - list of floats - helix sampling weights
       for each residue

       o *residue_strand_weights* - list of floats - strand sampling weights
       for each residue

       o *residue_coil_weights* - list of floats - coil sampling weights
       for each residue

       o *residue_turn1_weights* - list of floats - turn1 sampling weights
       for each residue

       o *residue_turn2_weights* - list of floats - turn2 sampling weights
       for each residue

       o *mesostates* - list of strings - allowed mesostates for each
       residue. *Note* - this attribute is present only if mesostates
       have been specified

    Methods

        o *get_atom* - get a reference to an atom in a molecule given
        the serial number of the residue it belongs to and the atoms'
        name

        o *update_coordinates* - recompute cartesian coordinates for
        a molecule after changes to the internal coordinates

        o *radius_of_gyration* - compute the radius of gyration of
        the molecule

        o *reset_conformation* - restore the conformation of the
        molecule to that specified in the input PDB file from which
        the molecule was first created
    """

    

    def __init__(self, filename):
        """New instance of LinusMol

        Arguments

            o *filename* - string - name of the PDB format file containing
            a description of the molecule

        Result

            New instace of LinusMol
        """
        
        import os
        self.filename = os.path.expanduser(filename)
        del os
        pdbio.protein_from_pdb(self)
        nr = self.num_residues
        self.residue_helix_weights = [0.25]*nr
        self.residue_strand_weights = [0.25]*nr
        self.residue_turn1_weights = [0.125]*nr
        self.residue_turn2_weights = [0.125]*nr
        self.residue_coil_weights = [0.125]*nr
        self.residue_PII_weights = [0.125]*nr
        fp = self._orig_fp_distance = []
        sp = self._orig_sp_angle = []
        tp = self._orig_tp_torsion = []
        for atom in self.atoms:
            fp.append(atom.fp_distance)
            sp.append(atom.sp_angle)
            tp.append(atom.tp_torsion)
        self.pmeso_grid = []
        for i in range(37):
            self.pmeso_grid.append(0.0)
        self.pmeso_weights = []
        for i in range(nr):
            self.pmeso_weights.append(self.pmeso_grid)



    def get_atom(self, resid, atomname):
        """Get a reference to an atom in a molecule

        Arguments

            o *resid* - integer - serial number of residue in which
            atom is present

            o *atomname* - string - name of atom

        Result

            Returns a reference to the atom.  If an atom with required
            name is not found in the resiude, *ValueError* is raised.
        """    
        i1 = self.residue_first_atom_indices[resid]
        i2 = self.residue_first_atom_indices[resid+1]
        for i in range(i1, i2):
            if self.atoms[i].name == atomname:
                return self.atoms[i]
        else:
            raise ValueError, 'No such atom: %s' % atomname
        

    def update_coordinates(self, start=None, end=None, ztox=Atom3d.ztox):
        """Compute cartesian coordinates from internal coordinates

        Arguments

            o *start* - integer - index of the first atom for which
            cartesian coordinates are to be computed.  This is an optional
            argument which defaults to the 4th atom in the molecule

            o *end* - integer - index of the last atom for which
            cartesian coordinates are to be computed.  This is an optional
            argument which defaults to the last atom in the molecule.
            *Note* - cartesian coordinates will be computed for all atoms
            for start upto and including end-1

        """
        
        if start == end == None:
            ztox(self.atoms)
        else:
            if start is None: start = 3
            if end is None: end = self.num_atoms
            ztox(self.atoms, start, end)

    def radius_of_gyration(self, start=None, end=None,
                                 rg=Atom3d.radius_of_gyration):
        """Radius of gyration of molecule or a subset of atoms in
        the molecule.  With no arguments specified the raidus of
        gyration of the entire molecule is computed.  If the
        radius of gyration of a contiguous subset of the atoms
        is required then the indices of the first and (last + 1)
        atoms should be supplied.

        Arguments

            o *start* - integer - index of the first atom

            o *end* - integer - index of last atom

        Result

            The radius of gyration as a floating point number
        """    
        if start == end == None:
            return rg(self.atoms)
        else:
            if start is None: start = 3
            if end is None: end = self.num_atoms
            return rg(self.atoms[start:end])
        
    def reset_conformation(self):
        """Restore the conformation of the molecule as existed in
        the PDB file from which the molecule was initially read.
        This function takes no arguments
        """
        
        pos = 0
        fp = self._orig_fp_distance
        sp = self._orig_sp_angle
        tp = self._orig_tp_torsion
        for atom in self.atoms:
            atom.fp_distance = fp[pos]
            atom.sp_angle = sp[pos]
            atom.tp_torsion = tp[pos]
            pos = pos + 1
        self.update_coordinates()


