"""Module implementing the scoring functions in LINUS

Functions in this module

    o *torsion_score* - compute torsion score of a molecule

    o *print_torsion_score* - print details of torsion score to file

    o *contact_score* - compute hydrophobic contact energy of a molecule

    o *print_contact_score* - print details of contact energy to file

    o *hbond_score* - compute hydrogen bond energy of a molecule

    o *print_hbond_score* - print details of hydrogen bond energy to
    file
"""

import Atom3d
import construct
import linusHbdSat
import sys
import operator
import math
from Numeric import *

try:
    from linusc import asa_evaluate
#   print 'Using C version'
except:
#   print 'Using python version'
    asa_evaluate = lambda a, b, c, d, e, f, g, h, i, j: 0.0

def torsion_score(mol, tp=1.0):
    """Torsion score of a molecule

    Arguments

        o *mol* - instance of LinusMol - molecule whose torsion
        score is to be computed

        o *tp* - float - value of torsion penalty

    Result

        The torsion score as a floating point number
    """
    
    rnames = mol.residue_names
    pos = 0
    etors = 0.0

    for atom in mol.phi_atoms:
        if 0.0 < atom.tp_torsion < 150.0:
            if rnames[pos] is 'GLY':
                etors = etors - tp
            elif rnames[pos] is 'ASN':
                etors = etors
            else:
                etors = etors + tp
        pos = pos + 1
    return etors

def print_torsion_score(mol, tp=1.0, file=sys.stdout):
    """List torsion score of a molecule to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose torsion
        score is to be computed

        o *tp* - float - value of torsion penalty

        o *file* - string or file object - either the name of a
        file or a file that has been previously opened for writing

    Result

        The torsion score as a floating point number
    """
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    ff.write('\n------ Torsion Penalty Listing -----\n')
    rnames = mol.residue_names
    pos = 0
    etors = 0.0

    for atom in mol.phi_atoms:
        phi = atom.tp_torsion
        if 0.0 < phi < 150.0:
            if  rnames[pos] is ('GLY'):
                etors = etors - tp
            else:
                etors = etors + tp
            ff.write('%d %s %.2f %.2f\n' % (pos+1, rnames[pos], phi, tp))
        pos = pos + 1
    ff.write('\nTotal torsion penalty = %.2f\n' % etors)
    if ff <> file: ff.close()
    return etors

        
def compaction_score(mol, cp=1.0, rg=Atom3d.radius_of_gyration,
                     exp=math.exp):
    return rg(mol.atoms)

def print_compaction_score(mol, cp=1.0, file=sys.stdout):
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    ff.write('\n------ Compaction Score -----\n')
    rg = Atom3d.radius_of_gyration(mol.atoms)
    ff.write('Radius of Gyration = %g\n' % rg)
    ff.write('\nCompaction Score = %g\n' % (math.exp(-rg*rg),))
    if ff <> file: ff.close()
    return rg*cp    

def contact_score(clist, cprobe):
    """Compute contact score for a molecule

    Arguments

        o *clist* - object - the object returned by the *make_contact_list*
        function in the *construct* module

        o *cprobe* - float - the distance over which the contact energy
        is to be scaled to zero.

    Result

        The contact energy as a floating point number
    """
    
    eborpc = 1.0/cprobe
    econ = 0.0
    for c1, c2, dmin, ene in clist:
        dmax = dmin + cprobe
        if c1.close(c2, dmax):
            d12 = c1.distance(c2)
##            if not c1.first_parent.close(c2.first_parent, d12):
            if d12 < dmin:
                econ = econ + ene
            else:
                econ = econ + ene*(dmax-d12)*eborpc
    return -econ

def print_contact_score(mol, clist, cprobe, file):
    """Print contact energy details to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose contact score
        is to be computed and listed to file

        o *clist* - object - the object returned by the *make_contact_list*
        function in the *construct* module

        o *cprobe* - float - the distance over which the contact energy
        is to be scaled to zero.

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The contact energy of the molecule
    """

    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file

    rnames = mol.residue_names
    atoms = mol.atoms
    fai = mol.residue_first_atom_indices

    Write = ff.write
    Write('\n------ Contact Score Listing -----\n')
    eborpc = 1.0/cprobe
    econ = 0.0
    pos = 0
    for c1, c2, dmin, ene in clist:
        dmax = dmin + cprobe
        if c1.close(c2, dmax):
            d12 = c1.distance(c2)
##            if not c1.first_parent.close(c2.first_parent, d12):
            if d12 < dmin:
                etmp = ene
            else:
                etmp = ene*(dmax-d12)*eborpc

            Write('%3d %s %s %3d %s %s Distance = %.3f Score = %.3f\n' % 
                   (c1.resnum+1, rnames[c1.resnum], c1.name,
                    c2.resnum+1, rnames[c2.resnum], c2.name,
                    d12, etmp))
            econ = econ + etmp
        pos = pos + 1
        
    Write('\nTotal Contact Score = %g\n' % -econ)

    if ff <> file:
        ff.close()
    return -econ

def hbond_score(hblist, ABS=abs, ANGLE=Atom3d.angle,
                        TORSION=Atom3d.torsion):
    """Compute hydrogen bond energy of a molecule

    Arguments

        o *hblist* - object - the object returned by the function
        *make_hbond_list* in the *construct* module

    Result

        The hydrogen bond energy as a floating point number
    """    
    ehb = 0.0
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close
        for acp, acp1, hbd, hbdmax, hbtors, hbene in alist:
            if Close(acp, hbdmax):
                if ABS(TORSION(acp, d1, da1, da2)) > hbtors and \
                   ANGLE(da1, d1, acp) > 90.0 and \
                   ANGLE(da2, d1, acp) > 90.0:
                    dist = d1.distance(acp)
                    if dist > hbd:
                        hbene = hbene * (hbdmax - dist)/(hbdmax - hbd)
                    ehb = ehb + hbene
                    break
    return -ehb

def print_hbond_score(mol, hblist, file, ABS=abs, ANGLE=Atom3d.angle,
                     TORSION=Atom3d.torsion):
    """List details of the hydrogen bond score to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose hydrogen bond
        energy is to be computed

        o *hblist* - object - the object returned by the function
        *make_hbond_list* in the *construct* module

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The hydrogen bond energy as a floating point number
    """    
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file

    Write = ff.write
    Write('\n------ Hydrogen Bond Score Listing -----\n')

    elist = []
    ehb = 0.0
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close
        for acp, acp1, hbd, hbdmax, hbtors, hbene in alist:
            if Close(acp, hbdmax) and \
               ANGLE(da1, d1, acp) > 90.0 and \
               ANGLE(da2, d1, acp) > 90.0 and \
               ABS(TORSION(acp, d1, da1, da2)) > hbtors:
                dist = d1.distance(acp)
                if dist > hbd:
                    hbene = hbene * (hbdmax - dist)/(hbdmax - hbd)
                ehb = ehb + hbene
                elist.append((d1, acp, dist, hbene))
                break

    rnames = mol.residue_names
    atoms = mol.atoms
    fai = mol.residue_first_atom_indices

    for rec in elist:
        d1, acp, dist, hbene = rec
        Write('%4d %3s %4s %4d %3s %4s Distance = %.3f Score = %.3f\n' %
              (d1.resnum+1, rnames[d1.resnum], d1.name, 
               acp.resnum+1, rnames[acp.resnum], acp.name, dist, hbene))
    Write('\nTotal Hydrogen Bond Score = %.3f\n' % -ehb)
    if ff <> file:
        ff.close()
    return -ehb

def solvation_score(atoms, wmax):
    """Compute the solvation score of a collection of atoms

    Arguments

        o *atoms* - list - a list of Atom3d instances

        o *wmax* - integer - maximum sequence separation between
        pairs of residues 

    Result

        The accessible surface area and solvation score as
        floating point numbers
    """    
    raise AttributeError, 'Function Not Implemented'

def print_solvation_score(p, cavpar, wmax, file):
    """List the solvation score of a molecule to file

    Arguments

        o *p* - instance of LinusMol - molecule whose solvation score
        is to be computed

        o *cavpar* - float - the cavitation parameter

        o *wmax* - integer - maximum sequence separation between a pair of
        residues

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The accessible surface area and solvation score as
        floating point numbers
    """
    
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    Write = ff.write
    Write('\n------ Solvation Score Listing -----\n')
    atoms = p.satoms
    area, esol = solvation_score(atoms, wmax)
    Write('Total Accessible Area = %g\n' % area)
    cavscr = area * cavpar
    Write('Cavitation Term = %g\n' % cavscr)
    for atom in atoms:
        Write('%s %6d Area=%12.6f Score=%12.6f\n' % (atom.name,
                                                     atom.resnum,
                                                     atom.sasa,
                                                     atom.sasa*atom.sasapar))
    s = cavscr + esol
    Write('Total Solvation Score = %g\n' % s)
    return s

   
def find_neighbors(atoms, coords, flags, probe, ext_rad,
         k, numatm, numext,nghlst):
    numngh = 0

    a = atoms[k]
    arad = a.radius + probe + probe
    
    for i in range(0,k):
        if not (flags[i]): continue
        b = atoms[i]
        d = a.distance(b)
        if d < (arad + b.radius):
#           print 'found neighbor ',atoms[i].name
            nghlst.append(i)

    for i in range(k+1,numatm):
        if not (flags[i]): continue
        b = atoms[i]
        d = a.distance(b)   
        if d < (arad + b.radius):
#           print 'found neighbor ',atoms[i].name
            nghlst.append(i)    

    numngh = len(nghlst)
    if not numext: return numngh

    ax = a.x
    ay = a.y
    az = a.z
    for i in range(numext):
        pos = i*3
        bx = coords[pos] - ax
        by = coords[pos+1] - ay
        bz = coords[pos+2] - az
        d = math.sqrt(bx*bx + by*by + bz*bz)
        if d < (arad + ext_rad):
            nghlst.append(-(i+1))

    numngh = len(nghlst)
    return numngh

def tri_buried(atoms, coords, tx, ty, tz, probe, ext_rad,
                         nghlst, numngh, nghstrt):
    for k in range(nghstrt,numngh):
        i = nghlst[k]
        if i >= 0:
            a = atoms[i]
            dx = a.x - tx
            dy = a.y - ty
            dz = a.z - tz
            d = dx*dx + dy*dy + dz*dz
            r = probe + a.radius
            if d < (r*r):
                nghstrt = k
                return 1,nghstrt

        else:
            pos = 3*(-(i+1))
            dx = coords[pos] - tx
            dy = coords[pos+1] - ty
            dz = coords[pos+2] - tz
            d = dx*dx + dy*dy + dz*dz
            r = ext_rad + probe
            if d < (r*r):
                nghstrt = k
                return 1,nghstrt

    for k in range(nghstrt):
        i = nghlst[k]
        if i >= 0:
            a = atoms[i]
            dx = a.x - tx
            dy = a.y - ty
            dz = a.z - tz
            d = dx*dx + dy*dy + dz*dz
            r = probe + a.radius
            if d < (r*r):
                nghstrt = k
                return 1,nghstrt

        else:
            pos = 3*(-(i+1))
            dx = coords[pos] - tx
            dy = coords[pos+1] - ty
            dz = coords[pos+2] - tz
            d = dx*dx + dy*dy + dz*dz
            r = ext_rad + probe
            if d < (r*r):
                nghstrt = k
                return 1,nghstrt

    return 0,nghstrt


def asa_evaluate(atoms, data, ext_coords, flags, probe,
                        ext_radius, use_data, use_ext, ndiv):
    from unit_triangles import triangles3
    triangles = triangles3
    ntrian = 960
    triarea = 4.0 * math.pi / ntrian
    numatm = len(atoms)
    if use_ext:
        numext = len(ext_coords)/3
    else:
        numext = 0
    PRESENT = 1
    CONTRIB = 2
    MAXNEIGHBORS = 200
    atot = 0.0
    for k in range(len(atoms)):
        nghlst = []
#       print atoms[k].name,flags[k],(flags[k] & CONTRIB)
#       if not ((flags[k] & CONTRIB) or use_data): 
        if not (flags[k]): 
#           print atoms[k].name
            continue
        a = atoms[k]
       
        numngh = find_neighbors(atoms, ext_coords, flags, probe, ext_radius,  
                 k, numatm, numext,nghlst)
#       sys.stdout.write('Atom %d %s has %d neighbors:\n' % (k,atoms[k].name,numngh))
#       for i in range(numngh):
#           sys.stdout.write('%s ' % atoms[nghlst[i]].name)
#       sys.stdout.write(' \n')
        nghstrt = 0
        nts = 0
        r = probe + a.radius
        x = a.x
        y = a.y
        z = a.z
        for i in range(ntrian):
            tx = triangles[i][0]*r + x
            ty = triangles[i][1]*r + y
            tz = triangles[i][2]*r + z
            buried,nghstrt = tri_buried(atoms, ext_coords, tx, ty, tz, probe, ext_radius,
                         nghlst, numngh, nghstrt)
            if not buried:
                nts = nts + 1
        area = nts * triarea * r * r 
        if use_data:
            data[k] = area
        
        if (flags[k] & CONTRIB):
            atot += area

    return atot


def print_hbs_casa(file, mol, flags, probe=1.4, ndiv=3,
                    ext_atoms=None, solv_list=None,ext_radius=1.25, title=''):
    """Prints PDB file containing:
           o number of solvating waters on backbone groups - occupancy column
           o CASA - Bfactor column
           o number of unsatisfied backbone Hbond D and A - in TER record

    Arguments

      o *filename* - filename or open file - location to which ASA
        information will be written

      o *mol* - instance of linusMol - molecule whose accessible surface
        area is to be calculated

      o *flags* - object - the object returned by the *make_asa_list*
        function in the **construct** module.

      o *probe* - float - the water probe size to use for accessible surface

      o *ndiv* - integer - a number from 1 to 5 representing the number of
        sampling points used in the area calculation.  From least (1) to
        most (5) accurate, this is either 60, 240, 960, 3840, or 15,350
        points.

      o *ext_atoms* - list of coordinates - a list of tuples, each
        tuple containing (x, y, z) of an additional atom to include
        in the ASA calculation

      o *ext_radius* - float - the radius of the atoms in ext_atoms.

      o *title* - string - An optional title for the PDB output
    """

    atoms = mol.atoms
    fai = mol.residue_first_atom_indices
    residue_names = mol.residue_names
    minres, maxres = construct.get_res_extents(mol)

  
    use_ext = 0
    ext_coords = None

    use_data = 1
    data = zeros(len(atoms), 'd')
  
    if ext_atoms:
        ext_coords = []
        map(lambda x: map(lambda y: ext_coords.append(y), x), ext_atoms)
        ext_coords = array(ext_coords, 'd')
        use_ext = len(ext_atoms)

    tot_asa =  asa_evaluate(atoms, data, ext_coords, flags, probe,
                            ext_radius, use_data, use_ext, ndiv)

    ff = file
    if type(file) == type('hi'):
        ff = open(file, 'wb')

    write = ff.write
    pdbfmt = 'ATOM  %5d %-4s %-3s  %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'

    p_solv_nrg = 0.0
    ap_solv_nrg = 0.0
    Gamma_p = 3.0/5.0
    Gamma_hb_oxy = 2.0
    Gamma_ap = 0.03
    CHASA = 0.0
    for i in xrange(minres,maxres):
        rname = residue_names[i]
        start = fai[i]
        end = fai[i+1]
        occ = 0.0
        for j in range(start, end):
            atom = atoms[j]
            residue_num = int(mol.res_pdb_number[atom.resnum])
            if atom.name == ' N  ':
                if solv_list[i][0][2] > 0:
                    p_solv_nrg = p_solv_nrg - (Gamma_p *(solv_list[i][0][2]))
                # if not solvated add penalty
                elif solv_list[i][0][2] < 0:
                    p_solv_nrg = p_solv_nrg + 1.0
    
                write(pdbfmt % (j+1, atom.name, rname, residue_num, atom.x, atom.y,
                            atom.z, solv_list[i][0][2],data[j]))
            elif atom.name == ' O  ':
                if solv_list[i][1][2] > 0:
                    if solv_list[i][1][3] == 0:
                        p_solv_nrg = p_solv_nrg - (Gamma_p *(solv_list[i][1][2]))
                    elif solv_list[i][1][3] > 0:
                        p_solv_nrg = p_solv_nrg - (Gamma_hb_oxy)
                elif solv_list[i][1][2] < 0:
                    p_solv_nrg = p_solv_nrg + 1.0
    
                write(pdbfmt % (j+1, atom.name, rname, residue_num, atom.x, atom.y,
                            atom.z, solv_list[i][1][2],data[j]))
            else:
                write(pdbfmt % (j+1, atom.name, rname, residue_num, atom.x, atom.y,
                            atom.z, occ,data[j]))
            if 'C' in atom.name:
                ap_solv_nrg = ap_solv_nrg + (Gamma_ap * data[j])
                CHASA = CHASA + data[j]

    aidx = j+2
    ridx = residue_num+1

    if ext_atoms:
        for x,y,z in ext_atoms:
            write(pdbfmt % (aidx, ' O  ', 'HOH', ridx, x, y, z, 0.0, 0.0))
            aidx = aidx + 1
            ridx = ridx + 1

    # note that this energy is not penalized by non-hbs donors or acceptors as in
    # the function below called calc_chasa_solv.
    tot_solv_nrg = ap_solv_nrg + p_solv_nrg
    write('TER %8.3f %8.3f\n' % (CHASA,tot_solv_nrg))

    if ff <> file:
        write('END\n')
        ff.close()

    return tot_asa,tot_solv_nrg


def calc_chasa_solv(mol,hblist):
    """Return the CHASA solvation energy for use in simulations.
       This adds a penalty of 5.0 for each non-hbs donor or acceptor
         unlike the print_hbs_casa function above.

    Arguments

      o *mol* - instance of linusMol - molecule whose accessible surface
        area is to be calculated
      o *hblist* - list of potential donors and acceptors
        In order for this list to be correct for the chasa_solv the
        following hbond parameters must be set in the command file:

#   sim.set_chasa_parameters(use_chasa=1)
#   # if chasa is used must use following
#   sim.set_hbond_parameters(use_hbond=1,
#                            hbond_score_short=?.?,
#                            hbond_score_long=?.?,
#                            hbond_winmax=NUMRES,
#                            use_sidechain_hbond=1)

    """

    atoms = mol.atoms
    fai = mol.residue_first_atom_indices
    residue_names = mol.residue_names
    minres, maxres = construct.get_res_extents(mol)


    use_ext = 0
    ext_coords = None

    use_data = 1
    data = zeros(len(atoms), 'd')

    numint_loos, numvirt_loos, bbtot, numbb,ext_atoms, solv_list = \
            linusHbdSat.mk_virt_bb_cntmult_loosHbd(mol,hblist)

    if ext_atoms:
        ext_coords = []
        map(lambda x: map(lambda y: ext_coords.append(y), x), ext_atoms)
        ext_coords = array(ext_coords, 'd')
        use_ext = len(ext_atoms)

    flags = construct.make_asa_list(mol)
    probe = 1.4
    ndiv = 3
    ext_radius = 1.25
    tot_asa =  asa_evaluate(atoms, data, ext_coords, flags, probe,
                            ext_radius, use_data, use_ext, ndiv)

    p_solv_nrg = 0.0
    ap_solv_nrg = 0.0
    Gamma_p = 3.0/5.0
    Gamma_hb_oxy = 2.0
    Gamma_ap = 0.03
    CHASA = 0.0
    non_hbd_score = 5.0
    for i in xrange(minres,maxres):
        rname = residue_names[i]
        start = fai[i]
        end = fai[i+1]
        occ = 0.0
        for j in range(start, end):
            atom = atoms[j]
            residue_num = int(mol.res_pdb_number[atom.resnum])
            if atom.name == ' N  ':
                if solv_list[i][0][2] > 0:
                    p_solv_nrg = p_solv_nrg - (Gamma_p *(solv_list[i][0][2]))
                elif solv_list[i][0][2] < 0:
                    p_solv_nrg = p_solv_nrg + non_hbd_score

            elif atom.name == ' O  ':
                if solv_list[i][1][2] > 0:
                    if solv_list[i][1][3] == 0:
                        p_solv_nrg = p_solv_nrg - (Gamma_p *(solv_list[i][1][2]))
                    elif solv_list[i][1][3] > 0:
                        p_solv_nrg = p_solv_nrg - (Gamma_hb_oxy)
                elif solv_list[i][1][2] < 0:
                    p_solv_nrg = p_solv_nrg + non_hbd_score

            elif 'C' in atom.name:
                ap_solv_nrg = ap_solv_nrg + (Gamma_ap * data[j])
                CHASA = CHASA + data[j]

    tot_solv_nrg = ap_solv_nrg + p_solv_nrg
#   print ap_solv_nrg,  p_solv_nrg

    return tot_solv_nrg


def calc_hbs_score_local(mol,hblist,res):
    """Return the local HBS score for use in simulations

    Arguments

      o *mol* - instance of linusMol - molecule whose accessible surface
        area is to be calculated
      o *hblist* - list of potential donors and acceptors
        In order for this list to be correct for the chasa_solv the
        following hbond parameters must be set in the command file:
      o *res* - the moveset anchor residue index
        HBS for the following residues is calculated,
        if res > 0:
            minres = res - 1
            maxres = res + 2
        bbtot          = number of total backbone N and O tested for hbs
        numbb          = number of backbone N and O that are either internally
                         hydrogen bonded or can be solvated
        NOTE: The radius of the water probe is decreased to 1.0 (rather than 1.25)
              to increase the acceptance ratio and to account for the rigidity of
              torsion angle moves.

    """

    numint_loos, numvirt_loos, bbtot, numbb,ext_atoms, solv_list = \
            linusHbdSat.mk_virt_bb_cntmult_loosHbd(mol,hblist,res,ext_rad=1.0)

    penalty = 5.0 * (bbtot - numbb)
    return penalty

try:    
    from linusc import *
except ImportError:
    print 'Using slower Python version of linusScore'
    
