from pylinus_1_0 import linusSim, linusScore, construct, linusHbdSat
from pylinus_1_0.Atom3d import ztox,angle,torsion
from pylinus_1_0.linusRes import RESIDUES
import sys, time
from math import sqrt, pi, sin, cos
import Numeric

DEGRAD = pi/180.0

def _norm(x1, y1, z1, x2, y2, z2, sqrt=sqrt):
    dx = x1 - x2; dy = y1 - y2; dz = z1 - z2;
    d = 1.0 / sqrt(dx*dx + dy*dy + dz*dz)
    return dx*d, dy*d, dz*d

def _ztox(coordsfp, coordssp, coordstp, zmvalu, zmindx):
    """Convert internal coordinates to cartesian coordinates

    Arguments

        o *coords* - list of type 'f' and 3 elements

        o *zmvalu* - list of type 'f' and 3 elements, where each
        element is respecively the distance, angle and torsion angle of
        the atom to it's parent, second parent and third parent.

        o *zmindx* - list of type 'i' and 3 elements, where each
        element is respectively the integer index of the atoms' first,
        second and third parents.

    Returns

        atcoor - the converted x,y,z coords
    """

    fp, sp, tp = zmindx
    fpd, spa, tpt = zmvalu

    spa = spa * DEGRAD
    tpt = tpt * DEGRAD

    sina = sin(spa)
    cosa = -cos(spa)

    sint = sina * sin(tpt)
    cost = sina * cos(tpt)

    fpx, fpy, fpz = coordsfp
    spx, spy, spz = coordssp
    tpx, tpy, tpz = coordstp

    u2x, u2y, u2z = _norm(fpx, fpy, fpz, spx, spy, spz)
    u1x, u1y, u1z = _norm(spx, spy, spz, tpx, tpy, tpz)

    cosine = u1x*u2x + u1y*u2y + u1z*u2z
    if abs(cosine) < 1.0:
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)

    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    dummy = Numeric.zeros((1, 3), 'f')
    atcoor = dummy[0]

    atcoor[0] = fpx + fpd*(u2x*cosa + u3x*sint + u4x)
    atcoor[1] = fpy + fpd*(u2y*cosa + u3y*sint + u4y)
    atcoor[2] = fpz + fpd*(u2z*cosa + u3z*sint + u4z)

    return atcoor

def get_atom_named(atoms, atom_name, start, end):
    for i in range(start, end):
        atom = atoms[i]
        if atom.name == atom_name:
            return atom

def get_scacceptor(atoms, start, end, res): 
    scatoms = []
    for arec in RESIDUES[res]:
        if arec[7] and arec[0] <> ' O  ':
            acp = (get_atom_named(atoms, arec[0], start, end))
            acp1 = acp.first_parent
            scatoms.append((acp,acp1))
    return scatoms

def make_hbtab_loos(p, hblist, ANGLE=angle,
             ABS=abs, TORSION=torsion):
    hbtab_loos = []
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close

        for acp, acp1, hbd, hbdmax, hbtors, hbene in alist:
            if d1.name == ' OG ' or d1.name == ' OG1' or d1.name == ' NZ ' \
                  or d1.name == ' OH ' or d1.name == ' NH1' or d1.name == ' NH2' \
                  or d1.name == ' ND1' or d1.name == ' ND2' or d1.name == ' OD2' \
                  or d1.name == ' SG ' or d1.name == ' NE2' or d1.name == ' OE2' \
                  or d1.name == ' NE1' or d1.name == ' OE1' or d1.name == ' OD1':
                if Close(acp, hbdmax):
                    if ANGLE(d1, acp, acp1) > 90.0:
                        dist = d1.distance(acp)
#                       print dist, d1.name, p.res_pdb_number[d1.resnum], acp.name, \
#                             p.res_pdb_number[acp.resnum]
                        hbtab_loos.append((d1, acp, dist))
            else:
                if Close(acp, hbdmax):
                    # first three keep oxygen in approximate +/- 70 deg cone of NH
                    # (it's actually a square)
                    # Equivalent to theta angle >= 110 in Kortemme et al. JMB, 2003
                    # Last angle keeps N in +/- 90 deg cone of CO
                    # Equivalent to psi angle >= 90 in Kortemme
                    if ABS(TORSION(acp, d1, da1, da2)) > 130.0 and \
                    ANGLE(da1, d1, acp) > 69.0 and \
                    ANGLE(da2, d1, acp) > 69.0 and \
                    ANGLE(d1, acp, acp1) > 90.0:
                        dist = d1.distance(acp)
#                       print dist, d1.name, p.res_pdb_number[d1.resnum], acp.name, \
#                             p.res_pdb_number[acp.resnum]
                        hbtab_loos.append((d1, acp, dist))

    return hbtab_loos

def _ishbonded_loos(p,atom, hbtab_loos,numint_loos):
    rn = p.residue_names
    rpn = p.res_pdb_number
    for rec in hbtab_loos:
        d1, acp, dist = rec
        if (atom.resnum==d1.resnum and \
            rn[atom.resnum]==rn[d1.resnum] and \
            atom.name==d1.name):
            numint_loos = numint_loos + 1
#           print 1, rpn[atom.resnum], rn[atom.resnum], atom.name,rpn[acp.resnum], rn[acp.resnum], acp.name,dist
            return 1,numint_loos
        elif (atom.resnum==acp.resnum and \
            rn[atom.resnum]==rn[acp.resnum] and \
            atom.name==acp.name):
            numint_loos = numint_loos + 1
#           print 2, atom.resnum+1, rn[atom.resnum], atom.name,d1.resnum+1, rn[d1.resnum], d1.name,dist
            return 1,numint_loos
    return 0,numint_loos


def _close(coords1, coords2, hsd):
    x1,y1,z1 = coords1
    x2,y2,z2 = coords2
    dist = sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
    if (dist < hsd):
        return 1

    return 0


#sys.stdout.write('Simulation starting time: %s\n' % time.ctime(time.time()))
Simobj = linusSim.Linus(sys.argv[1],None)
prot = Simobj.protein
prot.name = 'PROT'
numres = prot.num_residues
atoms = prot.atoms
rn = prot.residue_names      
fai = prot.residue_first_atom_indices

hbparms = Simobj.hbdpar
hbparms['use_hbond'] = 1
hbparms['use_sidechain_hbond'] = 1
wmin = 2
wmax = numres-1       
hblist = construct.make_hbond_list(prot, wmin, wmax, hbparms)

numint_loos, numvirt_loos, bbtot, numbb,wat_list, cntmult_list = linusHbdSat.mk_virt_bb_cntmult_loosHbd(prot,hblist,ext_rad=1.25)
bb_nonbd_loos = bbtot - numbb

sys.stdout.write('COMPND numint_loos numvirt_loos num_nonhbd_loos total_bb_polar\n')
sys.stdout.write('COMPND   %4d         %4d        %4d             %4d\n' % \
   (numint_loos, numvirt_loos,bb_nonbd_loos,bbtot))

asalist  = construct.make_asa_list(prot)
atot,tot_solv_nrg = linusScore.print_hbs_casa(sys.stdout, prot, 
    asalist, ext_atoms=wat_list,solv_list=cntmult_list)
#int_solv_nrg = tot_solv_nrg - (2.5 * numint_loos)
#print 'Internal Hbond + Solvation Energy = ', int_solv_nrg
#sys.stdout.write('Simulation Finished at: %s\n' % time.ctime(time.time()))
