""" Calculates hydrogen bond satisfaction.

Internal hydrogen bonded backbone N and O are identified using criteria
extracted from the PDB.  (See Kortemme et al., JMB, 2003) Sidechains are
included as possible donors and acceptors.

Each backbone N not hydrogen bonded will be solvated if sterically possible.
This is accomplished by attempting to place an oxygen (hard sphere radius,
ext_rad Ang.) proximate to the nitrogen at hydrogen bonding distance and
orientation. Five attempts are made at different postions within the cone
of approach. The number of successful placements is tracked.

Each backbone O whether or not it is internally hydrogen bonded is then
solvated, if possible, as described above for the N.

The module returns:

numint, numvirt, bbtot, numbb,wat_list, cntmult_list 

where,

numint         = number of internal hydrogen bonds
numvirt        = number of possible solvating waters (each backbone N
                 and O could have up to 5)
bbtot          = number of total backbone N and O (minus 1 because first
                 N is not looked at)
numbb          = number of backbone N and O that are either internally
                 hydrogen bonded or can be solvated
wat_list       = x,y,z coordinate list of solvating oxygens suitable for
                 passing to CHASA calculation
cntmult_list   = list of backbone N and O with two variables: number of
                 solvating waters (-1 if not solvated) and if an O whether
                 or not it is both hydrogen bonded and solvated. The
                 format is:

                 ([i,1,0,0],[i,2,0,0])

                 where first index is residue ID, second is N (1) or O (2), 
                 third is for number of solvating waters (or -1) and fourth
                 is flag for solvated O which is both internally hydrogen
                 bonded and solvated.
"""

import linusSim, linusScore, construct
from Atom3d import ztox,angle,torsion
from linusRes import RESIDUES
import sys
from math import sqrt, pi, sin, cos
import Numeric

DEGRAD = pi/180.0

def _norm(x1, y1, z1, x2, y2, z2, sqrt=sqrt):
    dx = x1 - x2; dy = y1 - y2; dz = z1 - z2;
    d = 1.0 / sqrt(dx*dx + dy*dy + dz*dz)
    return dx*d, dy*d, dz*d

def _ztox(coordsfp, coordssp, coordstp, zmvalu, zmindx):
    """Convert internal coordinates to cartesian coordinates

    Arguments

        o *coords* - list of type 'f' and 3 elements

        o *zmvalu* - list of type 'f' and 3 elements, where each
        element is respecively the distance, angle and torsion angle of
        the atom to it's parent, second parent and third parent.

        o *zmindx* - list of type 'i' and 3 elements, where each
        element is respectively the integer index of the atoms' first,
        second and third parents.

    Returns

        atcoor - the converted x,y,z coords
    """

    fp, sp, tp = zmindx
    fpd, spa, tpt = zmvalu

    spa = spa * DEGRAD
    tpt = tpt * DEGRAD

    sina = sin(spa)
    cosa = -cos(spa)

    sint = sina * sin(tpt)
    cost = sina * cos(tpt)

    fpx, fpy, fpz = coordsfp
    spx, spy, spz = coordssp
    tpx, tpy, tpz = coordstp

    u2x, u2y, u2z = _norm(fpx, fpy, fpz, spx, spy, spz)
    u1x, u1y, u1z = _norm(spx, spy, spz, tpx, tpy, tpz)

    cosine = u1x*u2x + u1y*u2y + u1z*u2z
    if abs(cosine) < 1.0:
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)

    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    dummy = Numeric.zeros((1, 3), 'f')
    atcoor = dummy[0]

    atcoor[0] = fpx + fpd*(u2x*cosa + u3x*sint + u4x)
    atcoor[1] = fpy + fpd*(u2y*cosa + u3y*sint + u4y)
    atcoor[2] = fpz + fpd*(u2z*cosa + u3z*sint + u4z)

    return atcoor

def get_atom_named(atoms, atom_name, start, end):
    for i in range(start, end):
        atom = atoms[i]
        if atom.name == atom_name:
            return atom

def get_scacceptor(atoms, start, end, res): 
    scatoms = []
    for arec in RESIDUES[res]:
        if arec[7] and arec[0] <> ' O  ':
            acp = (get_atom_named(atoms, arec[0], start, end))
            acp1 = acp.first_parent
            scatoms.append((acp,acp1))
    return scatoms

def make_hbtab_loos(p, hblist, ANGLE=angle,
             ABS=abs, TORSION=torsion):
    hbtab_loos = []
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close

        for acp, acp1, hbd, hbdmax, hbtors, hbene in alist:
            # for sidechains relax orientation of donor
            if (d1.name == ' OG ' or d1.name == ' OG1' or d1.name == ' NZ ' or
                d1.name == ' OH ' or d1.name == ' NH1' or d1.name == ' NH2' or
                d1.name == ' ND1' or d1.name == ' ND2' or d1.name == ' OD2' or
                d1.name == ' SG ' or d1.name == ' NE2' or d1.name == ' OE2' or
                d1.name == ' NE1' or d1.name == ' OE1' or d1.name == ' OD1'):
                if Close(acp, hbdmax):
                    if ABS(TORSION(acp, d1, da1, da2)) > 90.0 and \
                    ANGLE(da1, d1, acp) > 69.0 and \
                    ANGLE(da2, d1, acp) > 69.0 and \
                    ANGLE(d1, acp, acp1) > 90.0:
                        dist = d1.distance(acp)
#                       print dist, d1.name, p.res_pdb_number[d1.resnum], acp.name, \
#                             p.res_pdb_number[acp.resnum]
                        hbtab_loos.append((d1, acp, dist))
            elif (acp.name == ' OG ' or acp.name == ' OG1' or 
                acp.name == ' OH ' or 
                acp.name == ' ND1' or acp.name == ' ND2' or acp.name == ' OD2' or
                acp.name == ' SG ' or acp.name == ' NE2' or acp.name == ' OE2' or
                acp.name == ' NE1' or acp.name == ' OE1' or acp.name == ' OD1'):
                if Close(acp, hbdmax):
                    if ABS(TORSION(acp, d1, da1, da2)) > 90.0 and \
                    ANGLE(da1, d1, acp) > 69.0 and \
                    ANGLE(da2, d1, acp) > 69.0 and \
                    ANGLE(d1, acp, acp1) > 90.0:
                        dist = d1.distance(acp)   
#                       print dist, d1.name, da1.name, da2.name,p.res_pdb_number[d1.resnum], acp.name, acp1.name,\
#                             p.res_pdb_number[acp.resnum]
                        hbtab_loos.append((d1, acp, dist))
            else:
                if Close(acp, hbdmax):
                    # first three keep oxygen in approximate +/- 70 deg cone of NH
                    # (it's actually a square)
                    # Equivalent to theta angle >= 110 in Kortemme et al. JMB, 2003
                    # Last angle keeps N in +/- 90 deg cone of CO
                    # Equivalent to psi angle >= 90 in Kortemme
                    if ABS(TORSION(acp, d1, da1, da2)) > 130.0 and \
                    ANGLE(da1, d1, acp) > 69.0 and \
                    ANGLE(da2, d1, acp) > 69.0 and \
                    ANGLE(d1, acp, acp1) > 90.0:
                        dist = d1.distance(acp)
#                       print dist, d1.name, p.res_pdb_number[d1.resnum], acp.name, \
#                             p.res_pdb_number[acp.resnum]
                        hbtab_loos.append((d1, acp, dist))

    return hbtab_loos

def _ishbonded_loos(p,atom, hbtab_loos,numint_loos):
    rn = p.residue_names
    rpn = p.res_pdb_number
    for rec in hbtab_loos:
        d1, acp, dist = rec
        if (atom.resnum==d1.resnum and \
            rn[atom.resnum]==rn[d1.resnum] and \
            atom.name==d1.name):
            numint_loos = numint_loos + 1
#           print 1, rpn[atom.resnum], rn[atom.resnum], atom.name,rpn[acp.resnum], rn[acp.resnum], acp.name,dist
            return 1,numint_loos
        elif (atom.resnum==acp.resnum and \
            rn[atom.resnum]==rn[acp.resnum] and \
            atom.name==acp.name):
            numint_loos = numint_loos + 1
#           print 2, atom.resnum+1, rn[atom.resnum], atom.name,d1.resnum+1, rn[d1.resnum], d1.name,dist
            return 1,numint_loos
    return 0,numint_loos


def _close(coords1, coords2, hsd):
    x1,y1,z1 = coords1
    x2,y2,z2 = coords2
    dist = sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
    if (dist < hsd):
        return 1

    return 0

def mk_o2_solv(prot,atoms,start,end,acp,sp,tp,numvirt,
               tryall=0, watpdb=None, ext_rad=1.25):
    wat_coord = []
    #define a new single list zmatrix index
    ozmindx = [0,0,0]

    #use the O as the first parent     
    ozmindx[0] = acp      

    #use the C (same res) as the second parent
    ozmindx[1] = sp

    #use the CA (same res) as the third parent
    ozmindx[2] = tp

    #coords of the first parent, second and third    
    coordsfp = acp.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    #define the distance, angle, and torsion for the 180 deg "on top"
    #sol1
    ozmvalu = [2.95, 180.0, 0.0]

    owat2_crds = []
    owat2_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    gotone = 0
    scale = 0.9
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95
                # and H won't bump its own water of hydration
                hsd = 1.94
            else:      
                hsd = (ext_rad + atm_rad)

            if _close(owat2_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx2,yy2,zz2 = owat2_crds
        format16 = 'ATOM    900  O   HOH   921    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format16 % (xx2,yy2,zz2, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx2,yy2,zz2))
            gotone = 1
    elif watpdb and tryall:
        xx2,yy2,zz2 = owat2_crds
        format16 = 'ATOM    900  O   HOH   921    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format16 % (xx2,yy2,zz2, 0.0, 0.0))
        
    #If the 180 worked bail out and go to NH
#   if isbump < 1 :
#       return numvirt

    #define the distance, angle, and torsion for the "lower side"
    #sol2
    ozmvalu = [2.95, 130.0, 90.0]

    owat1_crds = []
    owat1_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat1_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx1,yy1,zz1 = owat1_crds
        format15 = 'ATOM    900  O   HOH   922    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format15 % (xx1,yy1,zz1, 0.0, 50.0))
            
        if gotone == 0:
            wat_coord.append((xx1,yy1,zz1))
            gotone = 1
    elif watpdb and tryall:
        xx1,yy1,zz1 = owat1_crds
        format15 = 'ATOM    900  O   HOH   922    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format15 % (xx1,yy1,zz1, 0.0, 0.0))

    #If this worked bail out and go to NH
#   if isbump < 1 :
#       return numvirt

    #define the distance, angle, and torsion for the "upper side"
    #sol3
    ozmvalu = [2.95, 130.0, -90.0]

    owat3_crds = []
    owat3_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat3_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx3,yy3,zz3 = owat3_crds
        format17 = 'ATOM    900  O   HOH   923    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format17 % (xx3,yy3,zz3, 0.0, 50.0))
        
        if gotone == 0:
            wat_coord.append((xx3,yy3,zz3))
            gotone = 1
    elif watpdb and tryall:
        xx3,yy3,zz3 = owat3_crds
        format17 = 'ATOM    900  O   HOH   923    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx3,yy3,zz3, 0.0, 0.0))


    #If this worked bail out and go to NH
#   if isbump < 1 :
#       return numvirt

    #define the distance, angle, and torsion for the near lone pair
    #sol4
    ozmvalu = [2.95, 130.0, 0.0]

    owat4_crds = []
    owat4_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat4_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx4,yy4,zz4 = owat4_crds
        format17 = 'ATOM    900  O   HOH   924    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format17 % (xx4,yy4,zz4, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx4,yy4,zz4))
            gotone = 1
    elif watpdb and tryall:
        xx4,yy4,zz4 = owat4_crds
        format17 = 'ATOM    900  O   HOH   924    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx4,yy4,zz4, 0.0, 0.0))


    #If this worked bail out and go to NH
#   if isbump < 1 :
#       return numvirt

    #define the distance, angle, and torsion for the far lone pair
    #sol5
    ozmvalu = [2.95, -130.0, 0.0]

    owat5_crds = []
    owat5_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat5_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx5,yy5,zz5 = owat5_crds
        format17 = 'ATOM    900  O   HOH   925    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format17 % (xx5,yy5,zz5, 0.0, 50.0))
        
        if gotone == 0:
            wat_coord.append((xx5,yy5,zz5))
            gotone = 1
    elif watpdb and tryall:
        xx5,yy5,zz5 = owat5_crds
        format17 = 'ATOM    900  O   HOH   925    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx5,yy5,zz5, 0.0, 0.0))
        
    return numvirt,wat_coord

def mk_nh_solv(prot,atoms,start,end,donor,sp,tp,numvirt,
               tryall=0, watpdb=None, ext_rad=1.25):
    wat_coord = []
    #define a new single list zmatrix index
    nzmindx = [0,0,0]        

    #use the N as the first parent
    nzmindx[0] = donor      

    #use the CA as the second parent
    nzmindx[1] = sp

    #use the C (prev res) as the third parent
    nzmindx[2] = tp

    #define the distance, angle, and torsion for 180 deg
    nzmvalu = [2.95, 119.0, 180.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    gotone = 0
    scale = 0.9
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))
            
        if gotone == 0:
            wat_coord.append((xx,yy,zz))
            gotone = 1
    elif tryall and watpdb:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))

#       return numvirt

    #define the distance, angle, and torsion for CA side
    nzmvalu = [2.95, 109.0, 180.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'):
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95
                # and H won't bump its own water of hydration
                hsd = 1.94
            else:
                hsd = (ext_rad + atm_rad)

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx,yy,zz))
            gotone = 1
    elif tryall and watpdb:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))
#       return numvirt

    #define the distance, angle, and torsion for C side
    nzmvalu = [2.95, 129.0, 180.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'):
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95
                # and H won't bump its own water of hydration
                hsd = 1.94
            else:
                hsd = (ext_rad + atm_rad)

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx,yy,zz))
            gotone = 1
    elif watpdb and tryall:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))
        
#       return numvirt

    #define the distance, angle, and torsion for right side
    nzmvalu = [2.95, 118.5, 170.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'):
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95
                # and H won't bump its own water of hydration
                hsd = 1.94
            else:
                hsd = (ext_rad + atm_rad)

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx,yy,zz))
            gotone = 1
    elif watpdb and tryall:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))
#       return numvirt

    #define the distance, angle, and torsion for left side
    nzmvalu = [2.95, 118.5, -170.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'):
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95
                # and H won't bump its own water of hydration
                hsd = 1.94
            else:
                hsd = (ext_rad + atm_rad)

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb and (tryall or not gotone):
            watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))

        if gotone == 0:
            wat_coord.append((xx,yy,zz))
            gotone = 1
    elif watpdb and tryall:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   910    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))
        
#       return numvirt

    return numvirt,wat_coord

def mk_amid_solv(prot,atoms,start,end,donor,sp,tp,numvirt,
                 tryall=0,watpdb=None, ext_rad=1.25):
    wat_coord = []
    #define a new single list zmatrix index
    nzmindx = [0,0,0]

    #use the N as the first parent
    nzmindx[0] = donor

    #use the C (prev res) as the second parent
    nzmindx[1] = sp

    #use the O (prev res) as the third parent          
    nzmindx[2] = tp

    #define the distance, angle, and torsion
    nzmvalu = [2.95, 180.0, 180.0]

    #coords of the first parent, second and third
    coordsfp = donor.coords()
    coordssp = sp.coords()
    coordstp = tp.coords() 
   
    nwat_crds = []
    nwat_crds = _ztox(coordsfp, coordssp, coordstp, nzmvalu, nzmindx)

    #check for bumps of N hbonded virtual O with all other atoms
    isbump = 0
    scale = 0.9
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'):
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(nwat_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break
        # if no bump write out virtual water
    if not isbump:
        numvirt =  numvirt + 1
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   911    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb: watpdb.write(format14 % (xx,yy,zz, 0.0, 50.0))
        wat_coord.append((xx,yy,zz))
    elif tryall and watpdb:
        xx,yy,zz = nwat_crds
        format14 = 'ATOM    900  O   HOH   911    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format14 % (xx,yy,zz, 0.0, 0.0))
        

    return numvirt,wat_coord

def mk_oh_solv(prot,atoms,start,end,acp,sp,tp,numvirt,tryall=0,watpdb=None,ext_rad=1.25):
    wat_coord = []
    #define a new single list zmatrix index
    ozmindx = [0,0,0]

    #use the O as the first parent 
    ozmindx[0] = acp

    #use the C (same res) as the second parent
    ozmindx[1] = sp

    #use the CA (same res) as the third parent
    ozmindx[2] = tp

    #coords of the first parent, second and third
    coordsfp = acp.coords()
    coordssp = sp.coords()
    coordstp = tp.coords()

    #define the distance, angle, and torsion for the "lower side"
    #sol2
    ozmvalu = [2.95, 130.0, 90.0]

    owat1_crds = []
    owat1_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    scale = 0.9
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat1_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water
    if not isbump:
        numvirt = numvirt + 1
        xx1,yy1,zz1 = owat1_crds
        format15 = 'ATOM    900  O   HOH   922    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb: watpdb.write(format15 % (xx1,yy1,zz1, 0.0, 50.0))
        wat_coord.append((xx1,yy1,zz1))
    elif tryall and watpdb:
        xx1,yy1,zz1 = owat1_crds
        format15 = 'ATOM    900  O   HOH   922    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format15 % (xx1,yy1,zz1, 0.0, 0.0))
        
    #If this worked bail out and go to NH
    if isbump < 1 and not tryall:
        return numvirt,wat_coord

    #define the distance, angle, and torsion for the "upper side"
    #sol3
    ozmvalu = [2.95, 130.0, -90.0]

    owat3_crds = []
    owat3_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat3_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water   
    if not isbump:  
        numvirt = numvirt + 1
        xx3,yy3,zz3 = owat3_crds
        format17 = 'ATOM    900  O   HOH   923    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb: watpdb.write(format17 % (xx3,yy3,zz3, 0.0, 50.0))
        wat_coord.append((xx3,yy3,zz3))
    elif tryall and watpdb:
        xx3,yy3,zz3 = owat3_crds
        format17 = 'ATOM    900  O   HOH   923    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx3,yy3,zz3, 0.0, 0.0))
 
    #If this worked bail out and go to NH
    if isbump < 1 and not tryall:
        return numvirt,wat_coord

    #define the distance, angle, and torsion for the near lone pair
    #sol4
    ozmvalu = [2.95, 130.0, 0.0]

    owat4_crds = []
    owat4_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat4_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water   
    if not isbump:  
        numvirt = numvirt + 1
        xx4,yy4,zz4 = owat4_crds
        format17 = 'ATOM    900  O   HOH   924    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb: watpdb.write(format17 % (xx4,yy4,zz4, 0.0, 50.0))
        wat_coord.append((xx4,yy4,zz4))
    elif tryall and watpdb:
        xx4,yy4,zz4 = owat4_crds
        format17 = 'ATOM    900  O   HOH   924    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx4,yy4,zz4, 0.0, 0.0))
        
    #If this worked bail out and go to NH
    if isbump < 1 and not tryall :
        return numvirt,wat_coord

    #define the distance, angle, and torsion for the far lone pair
    #sol5
    ozmvalu = [2.95, -130.0, 0.0]

    owat5_crds = []
    owat5_crds = _ztox(coordsfp, coordssp, coordstp, ozmvalu, ozmindx)

    # check for bumps of virt water with other atoms
    isbump = 0
    for atom in atoms:
        if (prot.residue_names[atom.resnum] != 'NME') and (prot.residue_names[atom.resnum] != 'ACE'): 
            atm_crds = atom.coords()
            atm_rad = scale * atom.radius
            if atom.name == ' H  ':
                # NH radius is 0.9, and H--O dist is 1.95        
                # and H won't bump its own water of hydration        
                hsd = 1.94
            else:        
                hsd = (ext_rad + atm_rad)        

            if _close(owat5_crds, atm_crds, hsd):
                #here only need the first instance of a bump
                isbump = 1
                break

    # if no bump write out virtual water   
    if not isbump:  
        numvirt = numvirt + 1
        xx5,yy5,zz5 = owat5_crds
        format17 = 'ATOM    900  O   HOH   925    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        if watpdb: watpdb.write(format17 % (xx5,yy5,zz5, 0.0, 50.0))
        wat_coord.append((xx5,yy5,zz5))
    elif tryall and watpdb:
        xx5,yy5,zz5 = owat5_crds
        format17 = 'ATOM    900  O   HOH   925    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
        watpdb.write(format17 % (xx5,yy5,zz5, 0.0, 0.0))

    return numvirt,wat_coord


def mk_virt_bb_cntmult_loosHbd(prot,hblist,res=0,tryall=0,watpdb=None,ext_rad=1.25):
    """
    Places oxygen in hydrogen bonding orientation to all donors and acceptors.
    Goes thru polypeptide and calls,
		mk_o2_solv, mk_nh_solv, mk_oh_solv, mk_amid_solv
    as appropriate.
    """

    numres = prot.num_residues
    atoms = prot.atoms
    rn = prot.residue_names
    fai = prot.residue_first_atom_indices

    minres,maxres = construct.get_res_extents(prot)

    hbtab_loos = make_hbtab_loos(prot, hblist)

    wat_list = []
    cntmult_list = []
    for i in range(minres,maxres+1):
        cntmult_list.append(([i,1,0,0],[i,2,0,0]))
    numbb = 0
    numposs = 0
    bbtot = 0
    numint_loos=0
    numint_sc_loos=0
    numvirt_loos=0
    numvirt_sc_loos=0
    # now see if only local score is requested
    if res > 0:
        minres = res - 1
        maxres = res + 2
    #backbone first
    for i in range(minres,maxres):
        start = fai[i]
        end = fai[i+1]
        acp = get_atom_named(atoms, ' O  ', start, end)
        sp = get_atom_named(atoms, ' C  ', start, end)
        tp = get_atom_named(atoms, ' CA ', start, end )

        bbtot = bbtot + 1
        numposs = numposs + 1

        hbond, numint_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_loos)

        old_numvirt_loos = numvirt_loos
        numvirt_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
             numvirt_loos,tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
        if len(wcoord) > 0:
            wat_list.append(wcoord[0])

        if hbond:
            numbb = numbb + 1
            if numvirt_loos > old_numvirt_loos:
                # if hbonded and solvent accessible just add one water
                cntmult_list[i][1][2] = 1
                # but flag this water as special for linusScore
                cntmult_list[i][1][3] = 1

        if not hbond:
            if numvirt_loos > old_numvirt_loos:
                numbb = numbb + 1
                num_solv = numvirt_loos - old_numvirt_loos
                cntmult_list[i][1][2] = num_solv
            else:
                cntmult_list[i][1][2] = -1

#               print prot.res_pdb_number[acp.resnum], prot.residue_names[acp.resnum] \
#                    , acp.name
 
    #construct a virtual water along the NH axis at 2.95 from the N
    for i in range(minres,maxres):
        if rn[i] <> 'PRO':
            if not ((i == 0) and (minres == 0)):
                start = fai[i]
                end = fai[i+1]
                prev = fai[i-1]      
                donor = get_atom_named(atoms, ' N  ', start, end)
                sp = get_atom_named(atoms, ' CA ', start, end)
                tp = get_atom_named(atoms, ' C  ', prev, start)
    
                bbtot = bbtot + 1
                numposs = numposs + 1
    
                hbond, numint_loos = _ishbonded_loos(prot, donor, hbtab_loos,numint_loos)
                if hbond:
                    numbb = numbb + 1
    
                if not hbond:
                    old_numvirt_loos = numvirt_loos
                    numvirt_loos,wcoord = mk_nh_solv(prot,atoms,start,end,donor,sp,tp, \
                     numvirt_loos,tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                    if len(wcoord) > 0:
                        wat_list.append(wcoord[0])
                    if numvirt_loos > old_numvirt_loos:
                        numbb = numbb + 1
                        num_solv = numvirt_loos - old_numvirt_loos
                        cntmult_list[i][0][2] = num_solv
    
                    else:
                        cntmult_list[i][0][2] = -1
#                   print prot.res_pdb_number[donor.resnum], prot.residue_names[donor.resnum] \
#                     , donor.name

    # now for sidechains
    for i in range(minres,maxres):
        if rn[i] == 'ARG':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the N
            acp = get_atom_named(atoms, ' NH1', start, end)
            sp = get_atom_named(atoms, ' CZ ', start, end)
            tp = get_atom_named(atoms, ' NE ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                   numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            #construct virtual waters at 2.95 from the N
            acp = get_atom_named(atoms, ' NH2', start, end)
            sp = get_atom_named(atoms, ' CZ ', start, end)
            tp = get_atom_named(atoms, ' NE ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # try all 4 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'ASN':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OD1', start, end)
            sp = get_atom_named(atoms, ' CG ', start, end)
            tp = get_atom_named(atoms, ' CB ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)
            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                   numvirt_sc_loos, tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            donor = get_atom_named(atoms, ' ND2', start, end)
            sp = get_atom_named(atoms, ' CG ', start, end)
            tp = get_atom_named(atoms, ' CB ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, donor, hbtab_loos,numint_sc_loos)

            if not hbond:
                numvirt_sc_loos,wcoord = mk_amid_solv(prot,atoms,start,end,donor,sp,tp, \
                    numvirt_sc_loos,tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'ASP':   
            start = fai[i] 
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OD1', start, end)
            sp = get_atom_named(atoms, ' CG ', start, end)
            tp = get_atom_named(atoms, ' CB ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                    numvirt_sc_loos,tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OD2', start, end)
            sp = get_atom_named(atoms, ' CG ', start, end)
            tp = get_atom_named(atoms, ' CB ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'GLN':   
            start = fai[i] 
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OE1', start, end)
            sp = get_atom_named(atoms, ' CD ', start, end)
            tp = get_atom_named(atoms, ' CG ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                   numvirt_sc_loos, tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            donor = get_atom_named(atoms, ' NE2', start, end)
            sp = get_atom_named(atoms, ' CD ', start, end)
            tp = get_atom_named(atoms, ' CG ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, donor, hbtab_loos,numint_sc_loos)

            if not hbond:
                numvirt_sc_loos,wcoord = mk_amid_solv(prot,atoms,start,end,donor,sp,tp, \
                      numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'GLU':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OE1', start, end)
            sp = get_atom_named(atoms, ' CD ', start, end)
            tp = get_atom_named(atoms, ' CG ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                   numvirt_sc_loos, tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OE2', start, end)
            sp = get_atom_named(atoms, ' CD ', start, end)
            tp = get_atom_named(atoms, ' CG ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)

            # if C=O not hbonded try all 5 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_o2_solv(prot,atoms,start,end,acp,sp,tp, \
                  numvirt_sc_loos, tryall=tryall, watpdb=watpdb, ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'HIS':  
            start = fai[i]
            end = fai[i+1]

            donor = get_atom_named(atoms, ' NE2', start, end)
            sp = get_atom_named(atoms, ' CD2', start, end)
            tp = get_atom_named(atoms, ' CG ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, donor, hbtab_loos,numint_sc_loos)
            if not hbond:
                numvirt_sc_loos,wcoord = mk_nh_solv(prot,atoms,start,end,donor,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

            donor = get_atom_named(atoms, ' ND1', start, end)
            sp = get_atom_named(atoms, ' CG ', start, end)
            tp = get_atom_named(atoms, ' CB ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, donor, hbtab_loos,numint_sc_loos)
            if not hbond:
                numvirt_sc_loos,wcoord = mk_nh_solv(prot,atoms,start,end,donor,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'LYS':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the N
            acp = get_atom_named(atoms, ' NZ ', start, end)
            sp = get_atom_named(atoms, ' CE ', start, end)
            tp = get_atom_named(atoms, ' CD ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)
            # try all 4 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'SER':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OG ', start, end)
            sp = get_atom_named(atoms, ' CB ', start, end)
            tp = get_atom_named(atoms, ' CA ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)
            # try all 4 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                   numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'THR':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OG1', start, end)
            sp = get_atom_named(atoms, ' CB ', start, end)
            tp = get_atom_named(atoms, ' CA ', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)
            # try all 4 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                     numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

        if rn[i] == 'TYR':
            start = fai[i]
            end = fai[i+1]

            #construct virtual waters at 2.95 from the O
            # one along the C=O axis
            acp = get_atom_named(atoms, ' OH ', start, end)
            sp = get_atom_named(atoms, ' CZ ', start, end)
            tp = get_atom_named(atoms, ' CE1', start, end )

            numposs = numposs + 1
            hbond, numint_sc_loos = _ishbonded_loos(prot, acp, hbtab_loos,numint_sc_loos)
            # try all 4 positions
            if not hbond:
                numvirt_sc_loos,wcoord = mk_oh_solv(prot,atoms,start,end,acp,sp,tp, \
                    numvirt_sc_loos, tryall=tryall, watpdb=watpdb,ext_rad=ext_rad)
                if len(wcoord) > 0:
                    wat_list.append(wcoord[0])

    total_loos = (numint_loos) + numvirt_loos
    bb_nonbd_loos = bbtot - numbb
    return numint_loos, numvirt_loos, bbtot, numbb,wat_list,cntmult_list

